import logging
from abc import ABCMeta, abstractmethod
from enum import IntEnum
from typing import Final, Mapping, Union

logger = logging.getLogger(__name__)


class LsiKindId(IntEnum):
    AD9082 = 1
    ADRF6780 = 2
    LMX2594 = 4
    AD5328 = 6
    GPIO = 7
    # Notes: the followings are only for QuEL-1 SE (SockClient doesn't support them)
    AD7490 = 16
    INA239 = 17
    MIXERBOARD_GPIO = 18
    PATHSELECTORBOARD_GPIO = 19
    POWERBOARD_PWM = 20

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_


class _ExstickgeProxyBase(metaclass=ABCMeta):
    _ADDR_MASKS: Final[Mapping[LsiKindId, int]] = {
        LsiKindId.AD9082: 0x7FFF,
        LsiKindId.ADRF6780: 0x003F,
        LsiKindId.LMX2594: 0x007F,
        LsiKindId.AD5328: 0x000F,
        LsiKindId.AD7490: 0x0000,
        LsiKindId.GPIO: 0x0000,
        LsiKindId.MIXERBOARD_GPIO: 0x0003,
        LsiKindId.PATHSELECTORBOARD_GPIO: 0x0000,
        LsiKindId.POWERBOARD_PWM: 0x007F,
    }

    def __init__(
        self,
        target_address: str,
        target_port: int,
        timeout: float,
    ):
        self._target = (target_address, target_port)
        self._timeout = timeout
        self._dump_enable = False

    def __del__(self):
        self.terminate()

    def dump_enable(self):
        self._dump_enable = True

    def dump_disable(self):
        self._dump_enable = False

    @abstractmethod
    def read_reg(self, kind, idx, addr) -> Union[int, None]:
        pass

    @abstractmethod
    def write_reg(self, kind, idx, addr, value) -> bool:
        pass

    @property
    def has_lock(self) -> bool:
        return False

    @abstractmethod
    def terminate(self):
        pass



from typing import Callable, TypeVar

from e7awghal import BoxLockDelegationError
from typing_extensions import ParamSpec

_trancated_traceback_enable: bool = True


class BoxLockError(Exception):
    pass


def set_trancated_traceback_for_lock_error(v: bool):
    global _trancated_traceback_enable
    _trancated_traceback_enable = v


Param = ParamSpec("Param")
RetType = TypeVar("RetType")


def guarded_by_box_lock(func: Callable[Param, RetType]) -> Callable[Param, RetType]:
    def wrapper(*args: Param.args, **kwargs: Param.kwargs) -> RetType:
        try:
            return func(*args, **kwargs)
        except BoxLockError as e:
            # Notes: the source of BoxLockError is apparent, it provides no valuable information.
            if _trancated_traceback_enable:
                raise e.with_traceback(None)
            else:
                raise e
        except BoxLockDelegationError as e:
            # Notes: the source of BoxLockDelegationError is apparent, it provides no valuable information.
            if _trancated_traceback_enable:
                raise BoxLockError(*e.args) from None
            else:
                raise e

    return wrapper



from quel_ic_config.exstickge_coap_client import AbstractSyncAsyncCoapClient
from quel_ic_config.exstickge_sock_client import AbstractLockKeeper


def force_unlock_all_boxes():
    AbstractLockKeeper.release_lock_all()
    AbstractSyncAsyncCoapClient.release_lock_all()



import logging
from typing import Any, Callable, Collection, Dict, Mapping, Set, Tuple, Union

from packaging.version import Version

from quel_ic_config.exstickge_coap_client import Quel1seBoard, _ExstickgeCoapClientBase
from quel_ic_config.exstickge_coap_tempctrl_client import _ExstickgeCoapClientQuel1seTempctrlBase
from quel_ic_config.exstickge_proxy import LsiKindId
from quel_ic_config.quel1_config_subsystem_tempctrl import Quel1seConfigSubsystemTempctrlDebugMixin
from quel_ic_config.quel1_thermistor import Quel1seOnboardThermistor, Quel1Thermistor
from quel_ic_config.quel1se_config_subsystem import _Quel1seConfigSubsystemBase
from quel_ic_config.quel_config_common import Quel1BoxType

logger = logging.getLogger(__name__)


class ExstickgeCoapClientAdda(_ExstickgeCoapClientQuel1seTempctrlBase):
    _VALID_BOXTYPE: Set[str] = {"quel1se-riken8", "quel1se-fujitsu11-a", "quel1se-fujitsu11-b"}

    _VERSION_SPEC: Tuple[Version, Version, Set[Version]] = Version("0.0.0"), Version("999.0.0"), set()

    _URI_MAPPINGS: Mapping[Tuple[LsiKindId, int], str] = {
        (LsiKindId.AD9082, 0): "adda/mxfe_0",
        (LsiKindId.AD9082, 1): "adda/mxfe_1",
        (LsiKindId.LMX2594, 0): "adda/pll_0",
        (LsiKindId.LMX2594, 1): "adda/pll_1",
        (LsiKindId.AD7490, 0): "tmp/ad_tc0",
    }

    _AVAILABLE_BOARDS: Tuple[Quel1seBoard, ...] = ()

    _TEMPCTRL_AD7490_NAME: Tuple[str, ...] = ("adda",)

    # Notes: no read is available for AD5328
    _READ_REG_PATHS: Mapping[LsiKindId, Callable[[int], str]] = {
        LsiKindId.AD9082: lambda addr: f"/reg/{addr:04x}",
        LsiKindId.LMX2594: lambda addr: f"/reg/{addr:04x}",
        LsiKindId.AD7490: lambda addr: "/ctrl",
    }

    _WRITE_REG_PATHS_AND_PAYLOADS: Mapping[LsiKindId, Callable[[int, int], Tuple[str, str]]] = {
        LsiKindId.AD9082: lambda addr, value: (f"/reg/{addr:04x}", f"{value:02x}"),
        LsiKindId.LMX2594: lambda addr, value: (f"/reg/{addr:04x}", f"{value:04x}"),
        LsiKindId.AD7490: lambda addr, value: ("/ctrl", f"{value:04x}"),
    }

    def __init__(
        self,
        target_addr: str,
        target_port: int = _ExstickgeCoapClientQuel1seTempctrlBase.DEFAULT_PORT,
        timeout: float = _ExstickgeCoapClientQuel1seTempctrlBase.DEFAULT_RESPONSE_TIMEOUT,
    ):
        super().__init__(target_addr, target_port, timeout)


class Quel1seAddaConfigSubsystem(_Quel1seConfigSubsystemBase, Quel1seConfigSubsystemTempctrlDebugMixin):
    __slots__ = ()

    _DEFAULT_CONFIG_JSONFILE = "quel-1se-adda.json"
    _NUM_IC: Dict[str, int] = {
        "ad9082": 2,
        "lmx2594": 2,
        "ad7490": 1,
        "powerboard_pwm": 0,
    }

    _PROXY_CLASSES: Tuple[type, ...] = (ExstickgeCoapClientAdda,)

    _GROUPS: Set[int] = {0, 1}

    _DAC_IDX: Dict[Tuple[int, int], Tuple[int, int]] = {
        (0, 0): (0, 0),
        (0, 1): (0, 1),
        (0, 2): (0, 2),
        (0, 3): (0, 3),
        (1, 0): (1, 3),
        (1, 1): (1, 2),
        (1, 2): (1, 1),
        (1, 3): (1, 0),
    }

    _ADC_IDX: Dict[Tuple[int, str], Tuple[int, int]] = {
        (0, "r"): (0, 3),
        (0, "m"): (0, 2),
        (1, "r"): (1, 3),
        (1, "m"): (1, 2),
    }

    _ADC_CH_IDX: Dict[Tuple[int, str], Tuple[Tuple[int, int], ...]] = {
        (0, "r"): ((0, 5),),
        (0, "m"): ((0, 4),),
        (1, "r"): ((1, 5),),
        (1, "m"): ((1, 4),),
    }

    _LO_IDX: Dict[Tuple[int, Union[int, str]], Tuple[int, int]] = {}
    _MIXER_IDX: Dict[Tuple[int, int], int] = {}
    _VATT_IDX: Dict[Tuple[int, int], Tuple[int, int]] = {}
    _RFSWITCH_IDX: Dict[Tuple[int, Union[int, str]], Tuple[int, int]] = {}
    _RFSWITCH_SUBORDINATE_OF: Dict[Tuple[int, Union[int, str]], Tuple[int, Union[int, str]]] = {}

    _DEFAULT_TEMPCTRL_AUTO_START_AT_LINKUP: bool = False

    _THERMISTORS: Dict[Tuple[int, int], Quel1Thermistor] = {
        (0, 0): Quel1seOnboardThermistor("adda_lmx2594_0"),
        (0, 1): Quel1seOnboardThermistor("adda_lmx2594_1"),
    }

    def __init__(
        self,
        css_addr: str,
        boxtype: Quel1BoxType,
        port: int = _ExstickgeCoapClientBase.DEFAULT_PORT,
        timeout: float = _ExstickgeCoapClientBase.DEFAULT_RESPONSE_TIMEOUT,
        sender_limit_by_binding: bool = False,
    ):
        _Quel1seConfigSubsystemBase.__init__(self, css_addr, boxtype, port, timeout, sender_limit_by_binding)
        self._construct_tempctrl_debug()

    def configure_peripherals(
        self,
        param: Dict[str, Any],
        *,
        ignore_access_failure_of_adrf6780: Union[Collection[int], None] = None,
        ignore_lock_failure_of_lmx2594: Union[Collection[int], None] = None,
    ) -> None:
        pass


class Quel2ProtoAddaConfigSubsystem(Quel1seAddaConfigSubsystem):
    __slots__ = ()

    _DEFAULT_CONFIG_JSONFILE = "quel-2-proto-adda.json"



import json
import os

from pydantic import BaseModel, Field

import quel_ic_config as qi


class Port(BaseModel):
    port: qi.Quel1PortType
    wait_ps_offset: int  # picosecond


class Box(BaseModel):
    name: str
    timecounter_offset: int = Field(default=0)
    wait_ps: int = Field(default=0)  # picosecond
    ports: list[Port]


class DeskewConfiguration(BaseModel):
    boxes: list[Box]


def load_default_configuration() -> DeskewConfiguration:
    xdg_config_home = os.environ.get("XDG_CONFIG_HOME", os.path.join(os.path.expanduser("~"), ".config"))

    config_dir = os.path.join(xdg_config_home, "quelware")
    config_file = os.path.join(config_dir, "deskew.json")

    if not os.path.exists(config_file):
        raise FileNotFoundError(f"Configuration file not found at {config_file}.")

    try:
        with open(config_file, "r") as f:
            obj = json.load(f)
    except Exception as e:
        raise ValueError(f"Error reading {config_file}: {e}") from e

    conf = DeskewConfiguration.model_validate(obj)
    return conf



import logging
from typing import Any, Collection, Dict, Set, Tuple, Union, cast

from quel_ic_config.exstickge_coap_client import _ExstickgeCoapClientBase, get_exstickge_server_info
from quel_ic_config.exstickge_proxy import LsiKindId
from quel_ic_config.quel1_config_subsystem_common import (
    Quel1ConfigSubsystemAd5328Mixin,
    Quel1ConfigSubsystemAd6780Mixin,
    Quel1ConfigSubsystemLmx2594Mixin,
    Quel1ConfigSubsystemMixerboardGpioMixin,
    Quel1ConfigSubsystemPathselectorboardGpioMixin,
    Quel1ConfigSubsystemRoot,
    Quel1GenericConfigSubsystemAd9082Mixin,
)
from quel_ic_config.quel_config_common import Quel1BoxType
from quel_ic_config.quel_ic import Ad9082Generic

logger = logging.getLogger(__name__)


class Ad9082Quel1se(Ad9082Generic):
    def _reset_pin_ctrl_cb(self, level: int) -> Tuple[bool]:
        proxy = cast(_ExstickgeCoapClientBase, self.proxy)
        proxy.write_reset(LsiKindId.AD9082, self.idx, level)
        return (True,)


class Quel1seConfigSubsystemAd9082Mixin(Quel1GenericConfigSubsystemAd9082Mixin):
    def _construct_ad9082(self):
        self._ad9082: tuple[Ad9082Generic, ...] = tuple(
            Ad9082Quel1se(self._proxy, idx) for idx in range(self._NUM_IC["ad9082"])
        )
        self.allow_dual_modulus_nco = True


class _Quel1seConfigSubsystemBase(
    Quel1ConfigSubsystemRoot,
    Quel1seConfigSubsystemAd9082Mixin,
    Quel1ConfigSubsystemLmx2594Mixin,
    Quel1ConfigSubsystemAd6780Mixin,
    Quel1ConfigSubsystemAd5328Mixin,
    Quel1ConfigSubsystemMixerboardGpioMixin,
    Quel1ConfigSubsystemPathselectorboardGpioMixin,
):
    __slots__ = ()

    _PROXY_CLASSES: Tuple[type, ...]

    _MXFE_IDXS: Set[int] = {0, 1}
    _LMX2594_OF_MXFES: Tuple[int, ...] = (0, 1)

    def __init__(
        self,
        css_addr: str,
        boxtype: Quel1BoxType,
        port: int = _ExstickgeCoapClientBase.DEFAULT_PORT,
        timeout: float = _ExstickgeCoapClientBase.DEFAULT_RESPONSE_TIMEOUT,
        sender_limit_by_binding: bool = False,
    ):
        Quel1ConfigSubsystemRoot.__init__(self, css_addr, boxtype, port, timeout, sender_limit_by_binding)

    def initialize(self) -> None:
        super().initialize()
        self._construct_ad9082()
        self._construct_lmx2594()
        self._construct_adrf6780()
        self._construct_ad5328()
        self._construct_mixerboard_gpio()
        self._construct_pathselectorboard_gpio()

    def _create_exstickge_proxy(
        self, port: int, timeout: float, sender_limit_by_binding: bool
    ) -> _ExstickgeCoapClientBase:
        # Notes: port will be available later.
        # Notes: sender_limit_by_binding may be available later.

        is_coap_firmware, coap_firmware_version, coap_boxtype = get_exstickge_server_info(self._css_addr)
        if is_coap_firmware:
            for proxy_cls in self._PROXY_CLASSES:
                if not issubclass(proxy_cls, _ExstickgeCoapClientBase):
                    raise AssertionError(f"invalid proxy class {proxy_cls.__name__} is in the list")
                if proxy_cls.matches(coap_boxtype, coap_firmware_version):
                    return proxy_cls(self._css_addr, port, timeout)
            else:
                raise RuntimeError(
                    f"unsupported CoAP firmware is running on {self._css_addr}: {coap_boxtype}:{coap_firmware_version}"
                )
        else:
            raise RuntimeError(f"CoAP server is not available on {self._css_addr}")

    def configure_peripherals(
        self,
        param: dict[str, Any],
        *,
        ignore_access_failure_of_adrf6780: Union[Collection[int], None] = None,
        ignore_lock_failure_of_lmx2594: Union[Collection[int], None] = None,
    ) -> None:
        if ignore_access_failure_of_adrf6780 is None:
            ignore_access_failure_of_adrf6780 = set()
        if ignore_lock_failure_of_lmx2594 is None:
            ignore_lock_failure_of_lmx2594 = set()

        # Notes: close all RF switches at first
        for i in range(self._NUM_IC["pathselectorboard_gpio"]):
            self.init_pathselectorboard_gpio(i, param["pathselectorboard_gpio"][i])

        # Notes: release reset of CPLDs on all the peripheral board
        proxy = cast(_ExstickgeCoapClientBase, self._proxy)
        for board in proxy.available_boards_with_cpld:
            if not proxy.read_board_active(board):
                logger.info(f"releasing reset of board '{board.value}'")
                proxy.write_board_active(board, True)
            else:
                logger.info(f"board '{board.value}' is already activated")

        # Notes: initialize ICs on mixer board 0 for RF
        for i in range(self._NUM_IC["ad5328"]):
            self.init_ad5328(i, param["ad5328"][i])

        for i in range(self._NUM_IC["adrf6780"]):
            proxy.write_reset(LsiKindId.ADRF6780, i, 1)
            self.init_adrf6780(i, param["adrf6780"][i], ignore_id_mismatch=i in ignore_access_failure_of_adrf6780)

        for i in range(2, self._NUM_IC["lmx2594"]):
            self.init_lmx2594(i, param["lmx2594"][i], ignore_lock_failure=i in ignore_lock_failure_of_lmx2594)

    def configure_all_mxfe_clocks(
        self, param: dict[str, Any], ignore_lock_failure_of_lmx2594: Union[Collection[int], None] = None
    ) -> None:
        if ignore_lock_failure_of_lmx2594 is None:
            ignore_lock_failure_of_lmx2594 = {}

        for group in range(2):
            lmx2594_idx = 0 + group
            self.init_lmx2594(
                lmx2594_idx,
                param["lmx2594"][lmx2594_idx],
                ignore_lock_failure=lmx2594_idx in ignore_lock_failure_of_lmx2594,
            )

    def configure_mxfe(
        self,
        mxfe_idx: int,
        param: Dict[str, Any],
        *,
        hard_reset: bool = False,
        soft_reset: bool = False,
        use_204b: bool = True,
        use_bg_cal: bool = False,
        ignore_crc_error: bool = False,
    ) -> bool:
        self._validate_group(mxfe_idx)

        self.ad9082[mxfe_idx].configure(
            param["ad9082"][mxfe_idx],
            hard_reset=hard_reset,
            soft_reset=soft_reset,
            use_204b=use_204b,
            use_bg_cal=use_bg_cal,
        )
        health, lglv, diag = self.check_link_status(mxfe_idx, True, ignore_crc_error)
        logger.log(lglv, diag)
        return health

    def reconnect_mxfe(
        self,
        mxfe_idx: int,
        *,
        ignore_crc_error: bool = False,
    ) -> bool:
        self._validate_mxfe(mxfe_idx)
        pll = self.lmx2594[self._LMX2594_OF_MXFES[mxfe_idx]]
        dr = pll.get_divider_ratio()[0]  # Notes: assuming that OUTA is used.
        se = pll.get_sync_enable()
        if dr == 0:
            raise RuntimeError(f"PLL of AD9082-#{mxfe_idx} is not activated")
        ref_clk = pll.get_lo_multiplier() * 100_000_000 // dr
        if se:
            ref_clk *= 4

        self.ad9082[mxfe_idx].reconnect(ref_clk)
        health, lglv, diag = self.check_link_status(mxfe_idx, False, ignore_crc_error)
        logger.log(lglv, diag)
        return health

    def dump_channel(self, group: int, line: int, channel: int) -> Dict[str, Any]:
        """dumping the current configuration of a transmitter channel.

        :param group: an index of a group which the channel belongs to.
        :param line: a group-local index of a line which the target channel belongs to.
        :param channel: a line-local index of the channel.
        :return: the current configuration information of the channel.
        """
        self._validate_channel(group, line, channel)
        return {
            "fnco_freq": self.get_dac_fnco(group, line, channel),
        }

    def dump_line(self, group: int, line: int) -> Dict[str, Any]:
        """dumping the current configuration of a transmitter line.

        :param group: an index of a group which the line belongs to.
        :param line: a group-local index of the line.
        :return: the current configuration information of the line.
        """
        self._validate_line(group, line)
        r: Dict[str, Any] = {}
        r["channels"] = {
            ch: self.dump_channel(group, line, ch) for ch in range(self.get_num_channels_of_line(group, line))
        }
        r["cnco_freq"] = self.get_dac_cnco(group, line)
        r["fullscale_current"] = self.get_fullscale_current(group, line)
        if (group, line) in self._LO_IDX:
            d_ratio = self.get_divider_ratio(group, line)
            if d_ratio > 0:
                r["lo_freq"] = self.get_lo_multiplier(group, line) * 100_000_000 // d_ratio
            else:
                r["lo_freq"] = 0
        if (group, line) in self._VATT_IDX:
            r["sideband"] = self.get_sideband(group, line)
            vatt = self.get_vatt_carboncopy(group, line)
            if vatt is not None:
                r["vatt"] = vatt
        if (group, line) in self._RFSWITCH_IDX:
            r["rfswitch"] = "block" if self.is_blocked_line(group, line) else "pass"
        return r

    def dump_rchannel(self, group: int, rline: str, rchannel: int) -> Dict[str, Any]:
        """dumping the current configuration of the receiver channel.

        :param group: an index of a group which the channel belongs to.
        :param rline: a group-local index of a line which the target channel belongs to.
        :param rchannel: a line-local index of the target channel.
        :return: the current configuration information of the channel.
        """
        self._validate_rchannel(group, rline, rchannel)
        return {
            "fnco_freq": self.get_adc_fnco(group, rline, rchannel),
        }

    def dump_rline(self, group: int, rline: str) -> Dict[str, Any]:
        """dumping the current configuration of a receiver line.

        :param group: an index of a group which the line belongs to.
        :param rline: a group-local index of the line.
        :return: the current configuration information of the line.
        """
        self._validate_rline(group, rline)
        r: Dict[str, Any] = {}
        if (group, rline) in self._LO_IDX:
            d_ratio = self.get_divider_ratio(group, rline)
            if d_ratio > 0:
                r["lo_freq"] = self.get_lo_multiplier(group, rline) * 100_000_000 // d_ratio
            else:
                r["lo_freq"] = 0
        r["cnco_freq"] = self.get_adc_cnco(group, rline)
        r["channels"] = {
            rch: self.dump_rchannel(group, rline, rch) for rch in range(self.get_num_rchannels_of_rline(group, rline))
        }
        if (group, rline) in self._RFSWITCH_IDX:
            r["rfswitch"] = "loop" if self.is_blocked_line(group, rline) else "open"
        return r



import logging
from dataclasses import dataclass, field
from typing import Dict, List, Union

from quel_ic_config.abstract_ic import AbstractIcMixin, AbstractIcReg, b_1bf_bool, b_nbf, p_1bf_bool, p_nbf

logger = logging.getLogger(__name__)


@dataclass
class Ad7490ConfigReg(AbstractIcReg):
    write: bool = field(default=True)  # [11+4]
    seq: bool = field(default=True)  # [10+4]
    addr: int = field(default=0xF)  # [9+4:6+4]
    pm: int = field(default=0x3)  # [5+4:4:4]
    shadow: bool = field(default=True)  # [3+4]
    weak: bool = field(default=False)  # [2+4]
    range: bool = field(default=True)  # [1+4], 0 for 2V_ref, 1 for V_ref
    coding: bool = field(default=True)  # [0+4], 0 for signed, 1 for unsigned

    def parse(self, v: int) -> None:
        self.write = p_1bf_bool(v, 15)
        self.seq = p_1bf_bool(v, 14)
        self.addr = p_nbf(v, 13, 10)
        self.pm = p_nbf(v, 9, 8)
        self.shadow = p_1bf_bool(v, 7)
        self.weak = p_1bf_bool(v, 6)
        self.range = p_1bf_bool(v, 5)
        self.coding = p_1bf_bool(v, 4)

    def build(self) -> int:
        return (
            b_1bf_bool(self.write, 15)
            | b_1bf_bool(self.seq, 14)
            | b_nbf(self.addr, 13, 10)
            | b_nbf(self.pm, 9, 8)
            | b_1bf_bool(self.shadow, 7)
            | b_1bf_bool(self.weak, 6)
            | b_1bf_bool(self.range, 5)
            | b_1bf_bool(self.coding, 4)
        )


class Ad7490Mixin(AbstractIcMixin):
    Regs: Dict[int, type] = {
        0: Ad7490ConfigReg,
    }
    RegNames: Dict[str, int] = {
        "Config": 0,
    }

    def __init__(self, name: str):
        super().__init__(name)
        config = Ad7490ConfigReg()
        self._default_config: int = config.build()
        self._num_read_iter: int = config.addr + 1

    def dump_regs(self) -> Dict[int, int]:
        raise NotImplementedError

    def set_default_config(self, **kwargs):
        config = Ad7490ConfigReg(**kwargs)
        self._default_config = config.build()
        self._num_read_iter = config.addr + 1

    def apply_default_config(self):
        # TODO: check the return value
        self.write_reg(0, self._default_config)

    def read_adcs_raw(self) -> Dict[int, int]:
        self.apply_default_config()
        adcs: Dict[int, int] = {}
        for i in range(self._num_read_iter):
            adcs[i] = self.read_reg(0)
        return adcs

    def read_adcs(self) -> Union[List[int], None]:
        self.apply_default_config()
        adcs: List[int] = []
        for i in range(self._num_read_iter):
            v = self.read_reg(0)
            if (v >> 12) & 0xF == i:
                adcs.append(v & 0xFFF)
            else:
                logger.error(f"broken data from {self.name} at channel {i}: {v:04x}")
                return None
        return adcs


# Notes: No helper class is implemented because AD7490 is simple enough.



import logging
import time
from concurrent.futures import Future, ThreadPoolExecutor
from typing import Any, Dict, List, Sequence, Set, Tuple, Union, cast

from quel_ic_config.exstickge_coap_tempctrl_client import Quel1seTempctrlState, _ExstickgeCoapClientQuel1seTempctrlBase
from quel_ic_config.quel1_config_subsystem_common import (
    Quel1ConfigSubsystemAd7490Mixin,
    Quel1ConfigSubsystemBaseSlot,
    Quel1ConfigSubsystemPowerboardPwmMixin,
)
from quel_ic_config.quel1_thermistor import Quel1Thermistor

logger = logging.getLogger(__name__)


class Quel1ConfigSubsystemTempctrlMixin(Quel1ConfigSubsystemBaseSlot):
    __slots__ = ()

    @property
    def tempctrl_auto_start_at_linkup(self):
        return False

    def get_tempctrl_state(self) -> Quel1seTempctrlState:
        # Notes: temperature control is running independently of the control subsystem.
        return Quel1seTempctrlState.RUN

    def get_tempctrl_state_count(self) -> int:
        # Notes: temperature control is running independently of the control subsystem.
        return 0

    def start_tempctrl(self, new_count: Union[int, None] = None) -> None:
        # Notes: this method should not be called, just exists for the purpose of static analysis.
        logger.warning(f"temperature control starts at power-on in this boxtype (={self._boxtype}), do nothing")


class Quel1seConfigSubsystemTempctrlMixin(Quel1ConfigSubsystemTempctrlMixin):
    __slots__ = ()

    _THERMISTORS: dict[tuple[int, int], Quel1Thermistor]
    _ACTUATORS: dict[str, tuple[str, int]]
    _DEFAULT_TEMPCTRL_AUTO_START_AT_LINKUP: bool
    _TEMPCTRL_POLLING_PERIOD: float = 0.5  # second
    _TEMPCTRL_POLLING_MAX_RETRY: int = 3  # times

    @classmethod
    def get_thermistor_desc(cls) -> dict[tuple[int, int], Quel1Thermistor]:
        return cls._THERMISTORS

    @classmethod
    def get_actuator_desc(cls) -> dict[str, tuple[str, int]]:
        return cls._ACTUATORS

    def _construct_tempctrl(self):
        self._tempctrl_auto_start_at_linkup: bool = self._DEFAULT_TEMPCTRL_AUTO_START_AT_LINKUP
        self._tempctrl_watcher = ThreadPoolExecutor(max_workers=1)

    def init_tempctrl(self, param: Dict[str, Any]):
        # Notes: currently nothing to do, define it for future.
        pass

    @property
    def tempctrl_auto_start_at_linkup(self):
        return self._tempctrl_auto_start_at_linkup

    @tempctrl_auto_start_at_linkup.setter
    def tempctrl_auto_start_at_linkup(self, v: bool):
        self._tempctrl_auto_start_at_linkup = v

    @property
    def tempctrl_loop_state(self) -> Dict[str, Union[int, str]]:
        proxy = cast(_ExstickgeCoapClientQuel1seTempctrlBase, self._proxy)
        loop_state: Quel1seTempctrlState = proxy.read_tempctrl_state()
        retval: Dict[str, Union[int, str]] = {
            "loop_count": proxy.read_tempctrl_loop_count(),
            "loop_state": loop_state.tostr(),
        }
        if loop_state == Quel1seTempctrlState.PRERUN:
            retval["time_to_start_running"] = proxy.read_tempctrl_state_count()
        return retval

    @property
    def tempctrl_error_count(self) -> Dict[str, int]:
        proxy = cast(_ExstickgeCoapClientQuel1seTempctrlBase, self._proxy)
        retval: Dict[str, int] = {
            "skip": proxy.read_tempctrl_skip_count(),  # Notes: number of skipped trials of thermal control
            "error": proxy.read_tempctrl_error_count(),  # Notes: SPI error
            "timeout": proxy.read_tempctrl_timeout_count(),  # Notes: SPI timeout
            "broken": proxy.read_tempctrl_broken_count(),  # Notes: broken channel index
            "mismatch": proxy.read_tempctrl_mismatch_count(),  # Notes: large mismatch between two consecutive readings
            "mismatch_max": proxy.read_tempctrl_mismatch_max(),
        }
        return retval

    def _convert_temperatures_index(self, tll: List[List[int]]) -> Dict[Tuple[int, int], float]:
        temperatures: Dict[Tuple[int, int], float] = {}

        for i in range(self._NUM_IC["ad7490"]):
            for j, t in enumerate(tll[i]):
                if (i, j) in self._THERMISTORS:
                    try:
                        temperatures[i, j] = self._THERMISTORS[i, j].convert(t)
                    except ValueError:
                        logger.error(f"unexpected reading {t} at AD7490[{i}], {j}-th input")

        return temperatures

    def _convert_temperatures_name(self, tll: List[List[int]]) -> Dict[str, float]:
        temperatures: Dict[str, float] = {}

        for i in range(self._NUM_IC["ad7490"]):
            for j, t in enumerate(tll[i]):
                if (i, j) in self._THERMISTORS:
                    try:
                        th = self._THERMISTORS[i, j]
                        temperatures[th.name] = th.convert(t)
                    except ValueError:
                        logger.error(f"unexpected reading {t} at AD7490[{i}], {j}-th input")

        return temperatures

    def _get_ad7490_readings_thread_main(self) -> List[List[int]]:
        proxy = cast(_ExstickgeCoapClientQuel1seTempctrlBase, self._proxy)
        cnt0 = proxy.read_tempctrl_loop_count()
        interval = proxy.read_tempctrl_loop_interval()
        num_iter = int(interval / self._TEMPCTRL_POLLING_PERIOD * 1.5) + 1
        for i in range(num_iter):
            cnt1 = proxy.read_tempctrl_loop_count()
            if cnt1 == cnt0:
                time.sleep(self._TEMPCTRL_POLLING_PERIOD)
            elif cnt1 == cnt0 + 1:
                break
            else:
                raise RuntimeError(f"an unexpected loop count {cnt1} comes after {cnt0}")
        else:
            raise TimeoutError(f"timeout occurred, loop count doesn't go up from {cnt0} too long")

        cnt2 = -1
        tll: List[List[int]] = []
        for _ in range(self._TEMPCTRL_POLLING_MAX_RETRY):
            # Notes: waiting for the firmware completes scanning temperatures.
            time.sleep(self._TEMPCTRL_POLLING_PERIOD)
            cnt2 = -1
            tll.clear()
            for i in range(self._NUM_IC["ad7490"]):
                cnt2, tl = proxy.read_tempctrl_rawdata_ad7490(i)
                if cnt2 != cnt1:
                    break
                tll.append(tl)
            if cnt2 != cnt1:
                logger.warning("the acquisition of temperature data is not completed yet")
                continue
            break
        else:
            raise RuntimeError(
                f"temperature data with an unexpected loop count {cnt2} (!= {cnt1}) is received "
                f"despite of {self._TEMPCTRL_POLLING_MAX_RETRY} trials"
            )
        return tll

    def get_tempctrl_temperature(self) -> Future[Dict[str, float]]:
        return self._tempctrl_watcher.submit(
            lambda: self._convert_temperatures_name(self._get_ad7490_readings_thread_main())
        )

    def get_tempctrl_temperature_now(self) -> Dict[str, float]:
        proxy = cast(_ExstickgeCoapClientQuel1seTempctrlBase, self._proxy)

        tll: List[List[int]] = []
        cnt0: int = -1
        cnt2: int = -1
        for j in range(2):
            cnt0 = proxy.read_tempctrl_loop_count()
            tll.clear()
            for i in range(self._NUM_IC["ad7490"]):
                cnt2, tl = proxy.read_tempctrl_rawdata_ad7490(i)
                if cnt0 != cnt2:
                    # Notes: retry from the first because jump over the tick, unfortunately.
                    break
                tll.append(tl)
            if len(tll) == self._NUM_IC["ad7490"]:
                break
        else:
            raise RuntimeError(f"temperature data with an unexpected loop count {cnt2} (!= {cnt0}) is received")

        return self._convert_temperatures_name(tll)

    def start_tempctrl(self, new_count: Union[int, None] = None) -> None:
        cur_state = self.get_tempctrl_state()
        if cur_state in {Quel1seTempctrlState.INIT, Quel1seTempctrlState.IDLE}:
            logger.info("starting temperature control")
        elif cur_state in {Quel1seTempctrlState.PRERUN, Quel1seTempctrlState.RUN, Quel1seTempctrlState.DRYRUN}:
            logger.info("re-starting temperature control")
        else:
            raise AssertionError
        proxy = cast(_ExstickgeCoapClientQuel1seTempctrlBase, self._proxy)
        proxy.write_tempctrl_state(Quel1seTempctrlState.PRERUN)
        if new_count is not None:
            proxy.write_tempctrl_state_count(new_count)

    def stop_tempctrl(self) -> None:
        cur_state = self.get_tempctrl_state()
        if cur_state == Quel1seTempctrlState.INIT:
            logger.info("re-initializing temperature control")
        elif cur_state in {Quel1seTempctrlState.PRERUN, Quel1seTempctrlState.RUN, Quel1seTempctrlState.DRYRUN}:
            logger.info("stopping and initializing temperature control")
        elif cur_state == Quel1seTempctrlState.IDLE:
            logger.info("initializing temperature control")
        else:
            raise AssertionError

        proxy = cast(_ExstickgeCoapClientQuel1seTempctrlBase, self._proxy)
        proxy.write_tempctrl_state(Quel1seTempctrlState.INIT)

    def start_tempctrl_external(self) -> None:
        logger.info("entering manual temperature control mode")
        proxy = cast(_ExstickgeCoapClientQuel1seTempctrlBase, self._proxy)
        proxy.write_tempctrl_state(Quel1seTempctrlState.DRYRUN)

    def get_tempctrl_state(self) -> Quel1seTempctrlState:
        proxy = cast(_ExstickgeCoapClientQuel1seTempctrlBase, self._proxy)
        return proxy.read_tempctrl_state()

    def set_tempctrl_state(self, new_state: Union[str, Quel1seTempctrlState]) -> None:
        if isinstance(new_state, str):
            new_state = Quel1seTempctrlState.fromstr(new_state)
        if not isinstance(new_state, Quel1seTempctrlState):
            raise TypeError(f"unexpected new_state: {new_state}")
        proxy = cast(_ExstickgeCoapClientQuel1seTempctrlBase, self._proxy)
        proxy.write_tempctrl_state(new_state)

    def get_tempctrl_state_count(self) -> int:
        proxy = cast(_ExstickgeCoapClientQuel1seTempctrlBase, self._proxy)
        return proxy.read_tempctrl_state_count()

    def set_tempctrl_state_count(self, new_count: int) -> None:
        proxy = cast(_ExstickgeCoapClientQuel1seTempctrlBase, self._proxy)
        return proxy.write_tempctrl_state_count(new_count)

    # TODO: fix it to return name-temp dictionary
    def get_tempctrl_setpoint(self) -> Dict[str, List[float]]:
        proxy = cast(_ExstickgeCoapClientQuel1seTempctrlBase, self._proxy)
        return proxy.read_tempctrl_feedback_setpoint()

    # TODO: fix it to take name-temp dictionary
    def set_tempctrl_setpoint(self, fan: Sequence[float], heater: Sequence[float]):
        proxy = cast(_ExstickgeCoapClientQuel1seTempctrlBase, self._proxy)
        proxy.write_tempctrl_feedback_setpoint(fan, heater)

    def get_tempctrl_gain(self) -> Dict[str, Dict[str, float]]:
        proxy = cast(_ExstickgeCoapClientQuel1seTempctrlBase, self._proxy)
        return proxy.read_tempctrl_feedback_coeffcient()

    def set_tempctrl_gain(
        self, fan: Union[Dict[str, float], None] = None, heater: Union[Dict[str, float], None] = None
    ) -> None:
        proxy = cast(_ExstickgeCoapClientQuel1seTempctrlBase, self._proxy)
        proxy.write_tempctrl_feedback_coeffcient(fan, heater)

    def get_tempctrl_actuator_output(self) -> Dict[str, Dict[str, float]]:
        proxy = cast(_ExstickgeCoapClientQuel1seTempctrlBase, self._proxy)
        outputs = proxy.read_tempctrl_rawdata_pwrpwm_current()
        fans: Dict[str, float] = {}
        heaters: Dict[str, float] = {}
        for name, (atype, aidx) in self.get_actuator_desc().items():
            if atype == "fan":
                fans[name] = outputs["fan"][aidx] / 1000
            elif atype == "heater":
                heaters[name] = outputs["heater"][aidx] / 1000
            else:
                raise AssertionError("never happens")
        return {
            "fan": fans,
            "heater": heaters,
        }

    def set_tempctrl_actuator_output(self, fan: Dict[str, float], heater: Dict[str, float]) -> None:
        fan_i: list[int] = [0 for _ in range(2)]
        heater_i: list[int] = [0 for _ in range(40)]
        for name, (atype, aidx) in self.get_actuator_desc().items():
            if atype == "fan":
                fan_i[aidx] = round(fan[name] * 1000)
            elif atype == "heater":
                heater_i[aidx] = round(heater[name] * 1000)
            else:
                raise AssertionError("never happens")
        proxy = cast(_ExstickgeCoapClientQuel1seTempctrlBase, self._proxy)
        proxy.write_tempctrl_rawdata_pwrpwm_next(fan_i, heater_i)


class Quel1seConfigSubsystemTempctrlDebugMixin(
    Quel1seConfigSubsystemTempctrlMixin, Quel1ConfigSubsystemAd7490Mixin, Quel1ConfigSubsystemPowerboardPwmMixin
):
    __slots__ = ()

    def _construct_tempctrl_debug(self):
        self._construct_tempctrl()
        self._construct_ad7490()
        self._construct_powerboard_pwm()

    def init_tempctrl_debug(self, param: Dict[str, Any]) -> None:
        self.init_tempctrl(param)
        for idx in range(self._NUM_IC["ad7490"]):
            self.init_ad7490(idx, param["ad7490"][idx])

    def _get_ad7490_readings_direct(self) -> List[List[int]]:
        tll: List[List[int]] = []

        for i in range(self._NUM_IC["ad7490"]):
            for k in range(3):
                tl = self.ad7490[i].read_adcs()
                if tl is not None:
                    break
            else:
                raise RuntimeError(f"failed to read temperature of {i}-th board repeatedly.")
            tll.append(tl)

        return tll

    def get_temperatures(self) -> Dict[Tuple[int, int], float]:
        return self._convert_temperatures_index(self._get_ad7490_readings_direct())

    def get_heater_master(self) -> bool:
        return self.powerboard_pwm[0].get_heater_master()

    def set_heater_master(self, en: bool) -> None:
        self.powerboard_pwm[0].set_heater_master(en)

    def get_heater_outputs(self, indices: Union[Set[int], None] = None) -> Dict[int, float]:
        ratios: Dict[int, float] = {}
        if indices is None:
            indices = {aidx for at, aidx in self.get_actuator_desc().values() if at == "heater"}
        for idx in range(self.powerboard_pwm[0].NUM_HEATER):
            if idx in indices:
                ratios[idx] = self.powerboard_pwm[0].get_heater_settings(idx)["high_ratio"]
        return ratios

    def set_heater_outputs(self, ratios: Dict[int, float]):
        indices = {aidx for at, aidx in self.get_actuator_desc().values() if at == "heater"}
        for idx, ratio in ratios.items():
            if idx not in indices:
                logger.warning(f"trying to configure an unavailable heater: {idx}")
            self.powerboard_pwm[0].set_heater_settings(idx, high_ratio=ratio)

    def get_fan_speed(self) -> float:
        s: List[float] = []
        for idx in {0, 1}:
            s.append(self.powerboard_pwm[0].get_fan_speed(idx))

        if s[0] == s[1]:
            return s[0]
        else:
            logger.warning(f"two fans have different speed: {s[0]:.3f} and {s[1]:.3f}, returns their average")
            return (s[0] + s[1]) / 2

    def set_fan_speed(self, ratio: float):
        for idx in {0, 1}:
            self.powerboard_pwm[0].set_fan_speed(idx, ratio)



from ._system_configuration import (
    Box,
    ClockMaster,
    SystemConfiguration,
    get_boxes,
    get_boxes_in_parallel,
    load_default_configuration,
    reconnect_and_get_link_status,
    reconnect_and_get_link_status_in_parallel,
)

__all__ = [
    "ClockMaster",
    "Box",
    "SystemConfiguration",
    "get_boxes",
    "get_boxes_in_parallel",
    "reconnect_and_get_link_status",
    "reconnect_and_get_link_status_in_parallel",
    "load_default_configuration",
]



import logging

from e7awghal.fwtype import E7FwType

logger = logging.getLogger(__name__)


class AbstractQuel1E7ResourceMapper:
    # Notes: (mxfe_idx, fduc_idx) --> awg_idx
    _AWGS_FROM_FDUC: dict[tuple[int, int], int]

    # Notes: (mxfe_idx, fddc_idx) --> cpmd_idx
    _CAPMOD_FROM_FDDC: dict[tuple[int, int], int]

    def __init__(self):
        return

    def get_awgs_of_mxfe(self, mxfe_idx: int) -> set[int]:
        return {a for (m, _), a in self._AWGS_FROM_FDUC.items() if m == mxfe_idx}

    def get_awg_from_fduc(self, mxfe_idx: int, fduc_idx: int) -> int:
        awg_idx: int = self._AWGS_FROM_FDUC.get((mxfe_idx, fduc_idx), -1)
        if awg_idx < 0:
            raise ValueError(f"no AWG is available for (mxfe_idx: {mxfe_idx}, fduc_idx: {fduc_idx})")
        return awg_idx

    def get_fduc_from_awg(self, awg_idx: int) -> tuple[int, int]:
        for mu, a in self._AWGS_FROM_FDUC.items():
            if a == awg_idx:
                return mu
        else:
            raise ValueError(f"invalid index of awg unit: {awg_idx}")

    def get_capmod_from_fddc(self, mxfe_idx: int, fddc_idx: int) -> int:
        cpmd_idx: int = self._CAPMOD_FROM_FDDC.get((mxfe_idx, fddc_idx), -1)
        if cpmd_idx < 0:
            raise ValueError(f"no AWG is available for (mxfe_idx: {mxfe_idx}, fddc_idx: {fddc_idx})")
        return cpmd_idx


class Quel1ConventionalE7ResourceMapper(AbstractQuel1E7ResourceMapper):
    # Notes: inverted order of AWG is caused by PCB pattern between FPGA and AD9082.
    #        (5th and 7th lanes are inverted for both AD9082s.)
    # TODO: should be fixed at AD9082 settings (and replace this class with Quel1CanonicalE7ResourceMapper)
    _AWGS_FROM_FDUC: dict[tuple[int, int], int] = {
        (0, 0): 15,
        (0, 1): 14,
        (0, 2): 13,
        (0, 3): 12,
        (0, 4): 11,
        (0, 5): 8,
        (0, 6): 9,
        (0, 7): 10,
        (1, 0): 7,
        (1, 1): 6,
        (1, 2): 5,
        (1, 3): 4,
        (1, 4): 3,
        (1, 5): 0,
        (1, 6): 1,
        (1, 7): 2,
    }

    # Notes: (mxfe_idx, fddc_idx) --> capmod_idx
    #        this is specific to version of e7awghw (feedback_early and simplemulti_standard)
    # TODO: the mixup between JESD204C TX lanes and FDDC should be filxed at AD9082 settings.
    #       (and replace this class with Quel1CanonicalE7ResourceMapper)
    _CAPMOD_FROM_FDDC: dict[tuple[int, int], int] = {
        (0, 5): 1,
        (0, 4): 3,
        (1, 5): 0,
        (1, 4): 2,
    }


def create_rmap_object(boxname: str, fw_type: E7FwType) -> AbstractQuel1E7ResourceMapper:
    if fw_type in {
        E7FwType.SIMPLEMULTI_CLASSIC,
        E7FwType.FEEDBACK_VERYEARLY,
    }:
        raise ValueError(f"firmware of the box '{boxname}' is deprecated (firmware type: {fw_type})")
    elif fw_type in {
        E7FwType.SIMPLEMULTI_STANDARD,
        E7FwType.NEC_EARLY,
        E7FwType.FEEDBACK_EARLY,
    }:
        return Quel1ConventionalE7ResourceMapper()
    else:
        raise ValueError(f"unexpceted firmware type: '{fw_type}'")


# Notes: this test is basically not necessary. Just for confirming deprecated settings are not applied by mistake.
def validate_configuration_integrity(
    convsel: list[int], fw_type: E7FwType, ignore_extraordinary_converter_select: bool = False
) -> None:
    if fw_type != E7FwType.SIMPLEMULTI_CLASSIC and (convsel[10] != 10 or convsel[11] != 11):
        if ignore_extraordinary_converter_select:
            logger.error("read-in and monitor-in ports are swapped unexpectedly")
        else:
            raise RuntimeError("read-in and monitor-in ports are swapped unexpectedly")



import logging
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from enum import IntEnum
from typing import Dict, Tuple, Union, cast

logger = logging.getLogger(__name__)


def p_1bf_bool(v: int, bidx: int) -> bool:
    return bool((v >> bidx) & 0x1)


def p_nbf(v: int, bidx0: int, bidx1: int) -> int:
    return (v >> bidx1) & ((1 << (bidx0 - bidx1 + 1)) - 1)


def b_1bf_bool(f: bool, bidx: int) -> int:
    return int(f) << bidx


def b_nbf(f: int, bidx0: int, bidx1: int) -> int:
    return (f & ((1 << (bidx0 - bidx1 + 1)) - 1)) << bidx1


class AbstractIcReg(metaclass=ABCMeta):
    @abstractmethod
    def parse(self, v: int) -> None:
        pass

    @abstractmethod
    def build(self) -> int:
        pass


class AbstractIcMixin(metaclass=ABCMeta):
    Regs: Dict[int, type] = {}
    RegNames: Dict[str, int] = {}

    def __init__(self, name: str):
        self.name = name

    def _read_and_parse_reg(self, regname: str) -> Tuple[int, AbstractIcReg]:
        addr = self.RegNames[regname]
        regcls = self.Regs[addr]
        reg = regcls()
        reg.parse(self.read_reg(addr))
        return addr, reg

    def _build_and_write_reg(self, addr: int, regobj: AbstractIcReg) -> None:
        self.write_reg(addr, regobj.build())

    @abstractmethod
    def dump_regs(self) -> Dict[int, int]:
        pass

    @abstractmethod
    def read_reg(self, addr: int) -> int:
        pass

    @abstractmethod
    def write_reg(self, addr: int, data: int) -> None:
        pass


class AbstractIcConfigHelper:
    def __init__(self, ic: AbstractIcMixin, no_read: bool = False):
        self.ic: AbstractIcMixin = ic
        self._no_read = no_read
        self.updated: Dict[int, int] = {}

    def dump_regs_parsed(self) -> Dict[int, AbstractIcReg]:
        parsed_dump: Dict[int, AbstractIcReg] = {}
        plain_dump = self.ic.dump_regs()
        for idx, regcls in self.ic.Regs.items():
            regobj = regcls()
            regobj.parse(plain_dump[idx])
            parsed_dump[idx] = regobj
        return parsed_dump

    def _parse_addr(self, addr: Union[str, int]) -> int:
        if isinstance(addr, str):
            return self.ic.RegNames[addr]
        elif isinstance(addr, int):
            return addr
        else:
            raise TypeError("illegal type of given address")

    def _zero_reg(self, addr: int, refer_to_cache: bool = False) -> AbstractIcReg:
        data = self.ic.Regs[addr]()
        if addr in self.updated and refer_to_cache:
            data.parse(self.updated[addr])
        else:
            data.parse(0)
        return cast(AbstractIcReg, data)

    def _read_reg(self, addr: int, refer_to_cache: bool = False) -> AbstractIcReg:
        data = self.ic.Regs[addr]()
        if addr in self.updated and refer_to_cache:
            data.parse(self.updated[addr])
        else:
            data.parse(self.ic.read_reg(addr))
        return cast(AbstractIcReg, data)

    def _write_reg(self, addr: int, data: Union[int, AbstractIcReg]) -> None:
        if isinstance(data, AbstractIcReg):
            if isinstance(data, self.ic.Regs[addr]):
                data = cast(AbstractIcReg, data).build()
            else:
                raise TypeError(f"mismatched register object for R{addr}")
        if not isinstance(data, int):
            raise TypeError("unexpected data is given")
        self.updated[addr] = data

    def read_reg(self, address: Union[int, str], refer_to_cache: bool = False) -> AbstractIcReg:
        """reading the value of a register at the address.

        :param address: the address of register to be read. you may use alias name instead.
        :param refer_to_cache: return the cached register value if True.
        :return: an instance of the corresponding register object
        """
        addr_ = self._parse_addr(address)
        return self._read_reg(addr_, refer_to_cache)

    def write_reg(self, address: Union[int, str], data: Union[int, AbstractIcReg]) -> None:
        """updating the value of a register into a cache. the cached values are written to the register when flush() is
        called.

        :param address: the address of a register to be updated.
        :param data: the value of the register.
        :return: None
        """
        addr_ = self._parse_addr(address)
        return self._write_reg(addr_, data)

    def write_field(self, addr: Union[int, str], **kwargs: Union[bool, int, IntEnum]):
        """a wrapper function of write_reg() providing a convenient way to update fields of a register.

        :param addr: the address of a register to be updated.
        :param kwargs: the values of fields of the register.
        :return: None
        """
        addr_ = self._parse_addr(addr)
        # TODO: consider the validity of this design carefully!
        if self._no_read:
            data: AbstractIcReg = self._zero_reg(addr_, True)
        else:
            data = self._read_reg(addr_, True)
        for k, v in kwargs.items():
            if hasattr(data, k):
                expected_type = type(data.__dict__[k])
                actual_type = type(v)
                if isinstance(v, expected_type):
                    data.__dict__[k] = v
                elif issubclass(expected_type, IntEnum):
                    data.__dict__[k] = expected_type(v)
                else:
                    raise TypeError(
                        f"unexpected value '{v}' for field '{k}', "
                        f"should be {expected_type} but type of the given value is {actual_type}"
                    )
            else:
                raise ValueError(f"invalid field name '{k}' for Reg:{addr}")
        self._write_reg(addr_, data)

    def discard(self):
        self.updated = {}

    @abstractmethod
    def flush(self, discard_after_flush=True):
        pass


@dataclass
class Gpio16(AbstractIcReg):
    b15: bool = field(default=False)
    b14: bool = field(default=False)
    b13: bool = field(default=False)
    b12: bool = field(default=False)
    b11: bool = field(default=False)
    b10: bool = field(default=False)
    b09: bool = field(default=False)
    b08: bool = field(default=False)
    b07: bool = field(default=False)
    b06: bool = field(default=False)
    b05: bool = field(default=False)
    b04: bool = field(default=False)
    b03: bool = field(default=False)
    b02: bool = field(default=False)
    b01: bool = field(default=False)
    b00: bool = field(default=False)

    def parse(self, v: int) -> None:
        self.b15 = p_1bf_bool(v, 15)
        self.b14 = p_1bf_bool(v, 14)
        self.b13 = p_1bf_bool(v, 13)
        self.b12 = p_1bf_bool(v, 12)
        self.b11 = p_1bf_bool(v, 11)
        self.b10 = p_1bf_bool(v, 10)
        self.b09 = p_1bf_bool(v, 9)
        self.b08 = p_1bf_bool(v, 8)
        self.b07 = p_1bf_bool(v, 7)
        self.b06 = p_1bf_bool(v, 6)
        self.b05 = p_1bf_bool(v, 5)
        self.b04 = p_1bf_bool(v, 4)
        self.b03 = p_1bf_bool(v, 3)
        self.b02 = p_1bf_bool(v, 2)
        self.b01 = p_1bf_bool(v, 1)
        self.b00 = p_1bf_bool(v, 0)

    def build(self) -> int:
        return (
            b_1bf_bool(self.b15, 15)
            | b_1bf_bool(self.b14, 14)
            | b_1bf_bool(self.b13, 13)
            | b_1bf_bool(self.b12, 12)
            | b_1bf_bool(self.b11, 11)
            | b_1bf_bool(self.b10, 10)
            | b_1bf_bool(self.b09, 9)
            | b_1bf_bool(self.b08, 8)
            | b_1bf_bool(self.b07, 7)
            | b_1bf_bool(self.b06, 6)
            | b_1bf_bool(self.b05, 5)
            | b_1bf_bool(self.b04, 4)
            | b_1bf_bool(self.b03, 3)
            | b_1bf_bool(self.b02, 2)
            | b_1bf_bool(self.b01, 1)
            | b_1bf_bool(self.b00, 0)
        )


@dataclass
class Gpio8(AbstractIcReg):
    b07: bool = field(default=False)
    b06: bool = field(default=False)
    b05: bool = field(default=False)
    b04: bool = field(default=False)
    b03: bool = field(default=False)
    b02: bool = field(default=False)
    b01: bool = field(default=False)
    b00: bool = field(default=False)

    def parse(self, v: int) -> None:
        self.b07 = p_1bf_bool(v, 7)
        self.b06 = p_1bf_bool(v, 6)
        self.b05 = p_1bf_bool(v, 5)
        self.b04 = p_1bf_bool(v, 4)
        self.b03 = p_1bf_bool(v, 3)
        self.b02 = p_1bf_bool(v, 2)
        self.b01 = p_1bf_bool(v, 1)
        self.b00 = p_1bf_bool(v, 0)

    def build(self) -> int:
        return (
            b_1bf_bool(self.b07, 7)
            | b_1bf_bool(self.b06, 6)
            | b_1bf_bool(self.b05, 5)
            | b_1bf_bool(self.b04, 4)
            | b_1bf_bool(self.b03, 3)
            | b_1bf_bool(self.b02, 2)
            | b_1bf_bool(self.b01, 1)
            | b_1bf_bool(self.b00, 0)
        )


@dataclass
class Gpio6(AbstractIcReg):
    b05: bool = field(default=True)
    b04: bool = field(default=True)
    b03: bool = field(default=True)
    b02: bool = field(default=True)
    b01: bool = field(default=True)
    b00: bool = field(default=True)

    def parse(self, v: int) -> None:
        self.b05 = p_1bf_bool(v, 5)
        self.b04 = p_1bf_bool(v, 4)
        self.b03 = p_1bf_bool(v, 3)
        self.b02 = p_1bf_bool(v, 2)
        self.b01 = p_1bf_bool(v, 1)
        self.b00 = p_1bf_bool(v, 0)

    def build(self) -> int:
        return (
            b_1bf_bool(self.b05, 5)
            | b_1bf_bool(self.b04, 4)
            | b_1bf_bool(self.b03, 3)
            | b_1bf_bool(self.b02, 2)
            | b_1bf_bool(self.b01, 1)
            | b_1bf_bool(self.b00, 0)
        )


@dataclass
class Gpio4(AbstractIcReg):
    b03: bool = field(default=False)
    b02: bool = field(default=False)
    b01: bool = field(default=False)
    b00: bool = field(default=False)

    def parse(self, v: int) -> None:
        self.b03 = p_1bf_bool(v, 3)
        self.b02 = p_1bf_bool(v, 2)
        self.b01 = p_1bf_bool(v, 1)
        self.b00 = p_1bf_bool(v, 0)

    def build(self) -> int:
        return b_1bf_bool(self.b03, 3) | b_1bf_bool(self.b02, 2) | b_1bf_bool(self.b01, 1) | b_1bf_bool(self.b00, 0)


@dataclass
class Gpio2(AbstractIcReg):
    b01: bool = field(default=False)
    b00: bool = field(default=False)

    def parse(self, v: int) -> None:
        self.b01 = p_1bf_bool(v, 1)
        self.b00 = p_1bf_bool(v, 0)

    def build(self) -> int:
        return b_1bf_bool(self.b01, 1) | b_1bf_bool(self.b00, 0)


@dataclass
class Uint16(AbstractIcReg):
    v: int = field(default=0)

    def parse(self, v: int) -> None:
        self.v = p_nbf(v, 15, 0)

    def build(self) -> int:
        return b_nbf(self.v, 15, 0)



import logging
from fractions import Fraction
from typing import Union

from pydantic import BaseModel, Field, model_validator
from typing_extensions import Self

logger = logging.getLogger(__name__)


class AbstractNcoFtw(BaseModel):
    # notes: x + a/b
    ftw: int = Field(ge=-0x8000_0000_0000, le=0x7FFF_FFFF_FFFF)
    delta_b: int = Field(ge=0x0000_0000_0000, le=0xFFFF_FFFF_FFFF)
    modulus_a: int = Field(ge=0x0000_0000_0000, le=0xFFFF_FFFF_FFFF)
    enable_fraction: bool = Field(default=True)

    @model_validator(mode="after")
    def check_numerator(self) -> Self:
        if self.enable_fraction:
            if self.delta_b >= self.modulus_a:
                raise ValueError("improper fraction is not allowed")
        else:
            if not (self.delta_b == 0 and self.modulus_a == 1):
                raise ValueError("delta_b and modulus_a must be 0 and 1 when fractional mode is disabled")
        return self

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, AbstractNcoFtw):
            return False
        return self.ftw == other.ftw and Fraction(self.delta_b, self.modulus_a) == Fraction(
            other.delta_b, other.modulus_a
        )

    def __repr__(self):
        return str(self)

    def __str__(self) -> str:
        return f"<ftw: {self.ftw} + {self.delta_b} / {self.modulus_a}>"

    @staticmethod
    def _encode_s48_as_u48(v: int):
        return v if v >= 0 else ((1 << 48) + v)

    @staticmethod
    def _decode_u48_as_s48(w: int):
        return w if w < (1 << 47) else w - (1 << 48)

    @staticmethod
    def _converter_frequency_as_interger(converter_freq_hz_: Union[int, float]) -> int:
        if isinstance(converter_freq_hz_, float):
            if not converter_freq_hz_.is_integer():
                raise ValueError("converter_freq_hz must be integer value")
            converter_freq_hz = int(converter_freq_hz_)
        elif isinstance(converter_freq_hz_, int):
            converter_freq_hz = converter_freq_hz_
        else:
            raise TypeError("illegal object for converter_freq_hz")

        return converter_freq_hz

    def to_frequency(self, converter_freq_hz_: Union[int, float]) -> float:
        converter_freq_hz = self._converter_frequency_as_interger(converter_freq_hz_)
        if self.enable_fraction:
            t = float((self.ftw + Fraction(self.delta_b, self.modulus_a)) * converter_freq_hz / (1 << 48))
        else:
            t = self.ftw * converter_freq_hz / (1 << 48)
        return t

    def truncate(self):
        self.delta_b = 0
        self.modulus_a = 1
        self.enable_fraction = False

    def round(self):
        frac = self.delta_b / self.modulus_a
        if (frac > 0.5) or (frac == 0.5 and self.ftw % 2 == 1):
            self.ftw += 1
        self.delta_b = 0
        self.modulus_a = 1
        self.enable_fraction = False

    def is_zero(self):
        return self.ftw == 0 and self.delta_b == 0

    def multiply(self, r: int) -> Self:
        x = self.ftw
        b = self.delta_b
        negative: bool = x < 0
        if negative:
            x = -x
            b = -b

        b *= r
        x = x * r + b // self.modulus_a
        b = b % self.modulus_a

        if negative:
            x = -x
            b = -b

        return self.__class__(ftw=x, delta_b=b, modulus_a=self.modulus_a)

    @classmethod
    def from_frequency(
        cls, nco_freq_hz_: Union[Fraction, float], converter_freq_hz_: Union[int, float], epsilon: float = 0.0
    ) -> Self:
        converter_freq_hz: int = cls._converter_frequency_as_interger(converter_freq_hz_)

        hcf: int = converter_freq_hz // 2
        if not (-hcf <= nco_freq_hz_ < hcf):
            raise ValueError(f"the given nco_frequency (= {nco_freq_hz_:f}Hz) is out of range")

        if not isinstance(converter_freq_hz, int):
            raise TypeError("converter_freq_hz must be integer")

        if isinstance(nco_freq_hz_, Fraction):
            nco_freq_hz: Fraction = nco_freq_hz_
        else:
            nco_freq_hz = Fraction(nco_freq_hz_)

        negative: bool = nco_freq_hz < 0
        if negative:
            t: Fraction = -nco_freq_hz * (1 << 48) / converter_freq_hz
        else:
            t = nco_freq_hz * (1 << 48) / converter_freq_hz
        t = t.limit_denominator(0xFFFF_FFFF_FFFF)

        x: int = int(t)
        if negative and t.denominator > 1:
            x += 1
        delta_b: int = t.numerator - t.denominator * x
        modulus_a: int = t.denominator
        if negative:
            x = -x
            delta_b = -delta_b  # Notes: delta_b becomes non-negative, finally.
        obj = cls(ftw=x, delta_b=delta_b, modulus_a=modulus_a)

        conv_error = abs(obj.to_frequency(converter_freq_hz) - nco_freq_hz_)
        if conv_error > epsilon:
            raise ValueError(
                f"large conversion error (= {conv_error}Hz) of the given nco frequency (= {nco_freq_hz:f}Hz)"
            )

        return obj



import logging
from abc import ABCMeta
from typing import Tuple, Union

from quel_ic_config.ad5328 import Ad5328Mixin
from quel_ic_config.ad7490 import Ad7490Mixin
from quel_ic_config.ad9082 import Ad9082Mixin
from quel_ic_config.adrf6780 import Adrf6780Mixin
from quel_ic_config.exstickge_proxy import LsiKindId, _ExstickgeProxyBase
from quel_ic_config.lmx2594 import Lmx2594Mixin
from quel_ic_config.mixerboard_gpio import MixerboardGpioMixin
from quel_ic_config.pathselectorboard_gpio import PathselectorboardGpioMixin
from quel_ic_config.powerboard_pwm import PowerboardPwmMixin
from quel_ic_config.rfswitcharray import (
    QubeRfSwitchArrayMixin,
    Quel1TypeARfSwitchArrayMixin,
    Quel1TypeBRfSwitchArrayMixin,
)

logger = logging.getLogger(__name__)


class Quel1Ic(metaclass=ABCMeta):
    def __init__(self, proxy: _ExstickgeProxyBase, kind: LsiKindId, idx: int):
        self.proxy = proxy
        self.kind = kind
        self.idx = idx
        # TODO: it may be fine if ExstickgeProxyBase conducts range check of (kind, idx) here.

    def _read_reg(self, addr: int) -> int:
        """A universal method for reading registers of the corresponding IC. You should not use this method
        if an equivalent specialized method is available.
        :param addr: the address of the register to read
        :return: the reading of the register
        """
        v = self.proxy.read_reg(self.kind, self.idx, addr)
        if v is None:
            raise RuntimeError(f"failed to read register 0x{addr:02x} of {self.kind.name}[{self.idx}]")
        return v & 0xFFFF

    def _read_reg_no_except(self, addr: int) -> Union[None, int]:
        """A universal method for reading registers of the corresponding IC. You should not use this method
        if an equivalent specialized method is available.
        :param addr: the address of the register to read
        :return: the reading of the register, it can be None for failure.
        """
        v = self.proxy.read_reg(self.kind, self.idx, addr)
        if v is None:
            return None
        else:
            return v & 0xFFFF

    def _write_reg(self, addr: int, data: int) -> None:
        """A universal method for writing registers of the corresponding IC. You should not use this method
        if an equivalent specialized method is available.
        :param addr: the address of the register to write
        :param data: a value to be written into the specified register
        """
        r = self.proxy.write_reg(self.kind, self.idx, addr, data & 0xFFFF)
        if not r:
            raise RuntimeError(
                f"failed to write 0x{data:04x} into register 0x{addr:02x} of {self.kind.name}[{self.idx}]"
            )
        return

    def _write_reg_no_except(self, addr: int, data: int) -> bool:
        """A universal method for writing registers of the corresponding IC. You should not use this method
        if an equivalent specialized method is available.
        :param addr: the address of the register to write
        :param data: a value to be written into the specified register
        :return: returns true if success
        """
        return self.proxy.write_reg(self.kind, self.idx, addr, data & 0xFFFF)


class Adrf6780(Adrf6780Mixin, Quel1Ic):
    def __init__(self, proxy: _ExstickgeProxyBase, idx: int):
        Adrf6780Mixin.__init__(self, f"ADRF6780[{idx}]")
        Quel1Ic.__init__(self, proxy, LsiKindId.ADRF6780, idx)

    def read_reg(self, addr: int) -> int:
        return Quel1Ic._read_reg(self, addr)

    def write_reg(self, addr: int, data: int) -> None:
        return Quel1Ic._write_reg(self, addr, data)


class Ad5328(Ad5328Mixin, Quel1Ic):
    def __init__(self, proxy: _ExstickgeProxyBase, idx: int):
        Ad5328Mixin.__init__(self, f"AD5328[{idx}]")
        Quel1Ic.__init__(self, proxy, LsiKindId.AD5328, idx)

    def read_reg(self, addr: int) -> int:
        return Quel1Ic._read_reg(self, addr)

    def write_reg(self, addr: int, data: int) -> None:
        self._carbon_copy_not_updated[addr] = data
        return Quel1Ic._write_reg(self, addr, data)


class Ad7490(Ad7490Mixin, Quel1Ic):
    def __init__(self, proxy: _ExstickgeProxyBase, idx: int):
        Ad7490Mixin.__init__(self, f"AD7490[{idx}]")
        Quel1Ic.__init__(self, proxy, LsiKindId.AD7490, idx)

    def read_reg(self, addr: int) -> int:
        return Quel1Ic._read_reg(self, addr)

    def write_reg(self, addr: int, data: int) -> None:
        return Quel1Ic._write_reg(self, addr, data)


# TODO: consider to subclass it into ForLo and ForRefClk or not.
class Lmx2594(Lmx2594Mixin, Quel1Ic):
    def __init__(self, proxy: _ExstickgeProxyBase, idx: int):
        Lmx2594Mixin.__init__(self, f"LMX2594[{idx}]")
        Quel1Ic.__init__(self, proxy, LsiKindId.LMX2594, idx)

    def read_reg(self, addr: int) -> int:
        return Quel1Ic._read_reg(self, addr)

    def write_reg(self, addr: int, data: int) -> None:
        return Quel1Ic._write_reg(self, addr, data)


class Ad9082Generic(Ad9082Mixin, Quel1Ic):
    def __init__(self, proxy: _ExstickgeProxyBase, idx: int):
        Ad9082Mixin.__init__(self, f"AD9082[{idx}]")
        Quel1Ic.__init__(self, proxy, LsiKindId.AD9082, idx)

    def _read_reg_cb(self, addr: int) -> Tuple[bool, int]:
        value = Quel1Ic._read_reg_no_except(self, addr)
        if value is None:
            logger.debug(f"failed to read from reg[{addr:04x}] of ad9082-#{self.idx:d}")
            return False, 0
        else:
            logger.debug(f"read {value:02x} from reg[{addr:04x}] of ad9082-#{self.idx:d}")
            return True, value

    def _write_reg_cb(self, addr: int, data: int) -> Tuple[bool]:
        logger.debug(f"write {data:02x} into reg[{addr:04x}] of ad9082-#{self.idx:d}")
        return (Quel1Ic._write_reg_no_except(self, addr, data),)


class QubeRfSwitchArray(QubeRfSwitchArrayMixin, Quel1Ic):
    def __init__(self, proxy: _ExstickgeProxyBase, idx: int):
        QubeRfSwitchArrayMixin.__init__(self, f"QubeRfSwitchArray[{idx}]")
        Quel1Ic.__init__(self, proxy, LsiKindId.GPIO, idx)

    def read_reg(self, addr: int) -> int:
        if addr != 0:
            raise ValueError(f"invalid address of {self.name}: {addr}")
        return Quel1Ic._read_reg(self, addr)

    def write_reg(self, addr: int, data: int) -> None:
        if addr != 0:
            raise ValueError(f"invalid address of {self.name}: {addr}")
        Quel1Ic._write_reg(self, addr, data)


class Quel1TypeARfSwitchArray(Quel1TypeARfSwitchArrayMixin, Quel1Ic):
    def __init__(self, proxy: _ExstickgeProxyBase, idx: int):
        Quel1TypeARfSwitchArrayMixin.__init__(self, f"Quel1TypeARfSwitchArray[{idx}]")
        Quel1Ic.__init__(self, proxy, LsiKindId.GPIO, idx)

    def read_reg(self, addr: int) -> int:
        if addr != 0:
            raise ValueError(f"invalid address of {self.name}: {addr}")
        return Quel1Ic._read_reg(self, addr)

    def write_reg(self, addr: int, data: int) -> None:
        if addr != 0:
            raise ValueError(f"invalid address of {self.name}: {addr}")
        Quel1Ic._write_reg(self, addr, data)


class Quel1TypeBRfSwitchArray(Quel1TypeBRfSwitchArrayMixin, Quel1Ic):
    def __init__(self, proxy: _ExstickgeProxyBase, idx: int):
        Quel1TypeBRfSwitchArrayMixin.__init__(self, f"Quel1TypeBRfSwitchArray[{idx}]")
        Quel1Ic.__init__(self, proxy, LsiKindId.GPIO, idx)

    def read_reg(self, addr: int) -> int:
        if addr != 0:
            raise ValueError(f"invalid address of {self.name}: {addr}")
        return Quel1Ic._read_reg(self, addr)

    def write_reg(self, addr: int, data: int) -> None:
        if addr != 0:
            raise ValueError(f"invalid address of {self.name}: {addr}")
        Quel1Ic._write_reg(self, addr, data)


class MixerboardGpio(MixerboardGpioMixin, Quel1Ic):
    def __init__(self, proxy: _ExstickgeProxyBase, idx: int):
        MixerboardGpioMixin.__init__(self, f"MixerboardGpio[{idx}]")
        Quel1Ic.__init__(self, proxy, LsiKindId.MIXERBOARD_GPIO, idx)

    def read_reg(self, addr: int) -> int:
        return Quel1Ic._read_reg(self, addr)

    def write_reg(self, addr: int, data: int) -> None:
        return Quel1Ic._write_reg(self, addr, data)


class PathselectorboardGpio(PathselectorboardGpioMixin, Quel1Ic):
    def __init__(self, proxy: _ExstickgeProxyBase, idx: int):
        PathselectorboardGpioMixin.__init__(self, f"PathselectorboardGpio[{idx}]")
        Quel1Ic.__init__(self, proxy, LsiKindId.PATHSELECTORBOARD_GPIO, idx)

    def read_reg(self, addr: int) -> int:
        return Quel1Ic._read_reg(self, addr)

    def write_reg(self, addr: int, data: int) -> None:
        return Quel1Ic._write_reg(self, addr, data)


class PowerboardPwm(PowerboardPwmMixin, Quel1Ic):
    def __init__(self, proxy: _ExstickgeProxyBase, idx: int):
        PowerboardPwmMixin.__init__(self, f"PowerboardPwm[{idx}]")
        Quel1Ic.__init__(self, proxy, LsiKindId.POWERBOARD_PWM, idx)

    def read_reg(self, addr: int) -> int:
        return Quel1Ic._read_reg(self, addr)

    def write_reg(self, addr: int, data: int) -> None:
        return Quel1Ic._write_reg(self, addr, data)



import logging
from typing import Dict

from quel_ic_config.abstract_ic import AbstractIcConfigHelper, AbstractIcMixin, Gpio4, Gpio16

logger = logging.getLogger(__name__)


MixerboardGpioRegs: Dict[int, type] = {
    0: Gpio16,
    1: Gpio16,
    2: Gpio4,
    3: Gpio4,
}


MixerboardGpioRegNames: Dict[str, int] = {
    "GPO_Enable": 0,
    "GPIO": 1,
    "ADRF6780_RSTn": 2,
    "ADRF6780_PWDN": 3,
}


class MixerboardGpioMixin(AbstractIcMixin):
    Regs = MixerboardGpioRegs
    RegNames = MixerboardGpioRegNames

    def __init__(self, name):
        super().__init__(name)

    def dump_regs(self) -> Dict[int, int]:
        """dump all the available registers.
        :return: a mapping between an address and a value of the registers
        """
        regs = {}
        for addr in self.Regs:
            regs[addr] = self.read_reg(addr)
        return regs


class MixerboardGpioConfigHelper(AbstractIcConfigHelper):
    """Helper class for programming GPIO with convenient notations. It also provides caching capability that
    keep modifications on the registers to write them at once with flash_updated() in a right order.
    """

    def __init__(self, ic: MixerboardGpioMixin):
        super().__init__(ic)

    def flush(self, discard_after_flush=True):
        for addr in self.ic.Regs:
            if addr in self.updated:
                self.ic.write_reg(addr, self.updated[addr])
        if discard_after_flush:
            self.discard()
